{****************************************************************************}
{                            Data Master 2000                                }
{****************************************************************************}
unit Process;

interface

uses Dialogs, SysUtils, MainWnd, DataWnd, CalcDlg, Common, Data, Strings,
  Forms, SelDlg, Controls;

type
  TOperationClass=class of TOperation;
  TOperation=class(TObject)        {models abstract "local" process operation}
  private
    Action: TObject;           // buffer for TAction object for SaveUndoInfo()
  public
    Form: TDataForm;                      {active form and its data container}
    Data: TContainer;
    XCol, YCol, BegLine, EndLine: integer;               {worksheet selection}
    Modified: boolean;            {clear in processitem if data not modified!}
    InsCol: Boolean;   {if true, place result to inserted column instead of Y}
    constructor Create(F: TDataForm; Op: TObject);
    procedure Execute;
    procedure ProcessItem(I:integer; D:TRealData); virtual; abstract;
  end;

  TSwapColumns=class(TOperation)
  public
    procedure ProcessItem(I:integer; D:TRealData); override;
  end;

  TParse=class(TOperation)
  public
    Macro: string;
    procedure ProcessItem(I:integer; D:TRealData); override;
  end;

  TIntegration=class(TOperation)                          {calculate integral}
  public
    procedure ProcessItem(I:integer; D:TRealData); override;
  end;

  TInterpolation=class(TOperation)           {add linearly interpolated curve}
  public
    Step: TReal;
    procedure ProcessItem(I:integer; D:TRealData); override;
  end;

  TDifferentiation=class(TOperation)       {3-point parabolic differentiation}
  public
    procedure ProcessItem(I:integer; D:TRealData); override;
  end;

  TSmooth=class(TOperation)               {use "best fit" method of smoothing}
  public
    HW: integer;           {initial halfwidth of vicinity for making best fit}
    Derive: boolean;              {set this flag=true to calculate derivative}
    procedure ProcessItem(I:integer; D:TRealData); override;
  end;

implementation

{ TOperation }

constructor TOperation.Create(F: TDataForm; Op: TObject);
begin Form:=F; Modified:=true; InsCol:=MainForm.Insert; Action:=Op; end;

procedure TOperation.Execute;
var I,J: integer; D: TRealData; S: string;
begin
  // if MainForm.GetActiveCtrlType=actPlot then               <- INCORRECT!!!!
  with MainForm.ActiveDataForm do if ActiveControl=PlotPanel then
  begin
    S:='';
    if ((Form.Plot.ThisSerie.XExpression<>'') or
        (Form.Plot.ThisSerie.YExpression<>'')) then S:=strSerieEmptyExpr;
    if ((Form.Plot.XAxis.Expression<>'') or (Form.Plot.YAxis.Expression<>''))
    then S:=S+CRLF+strAxisEmptyExpr;
    if (S<>'') and (MessageDlg(S, mtWarning, [mbOk,mbCancel], 0)=mrCancel)
    then Abort;           // warn user that curve X,Y and block X,Y may differ
    Data:=Form.Plot.ThisSerie.Container;
    BegLine:=Form.Plot.ThisSerie.FirstLine;
    EndLine:=Form.Plot.ThisSerie.LastLine;
    XCol:=Form.Plot.ThisSerie.XColumn;
    YCol:=Form.Plot.ThisSerie.YColumn;
    Form:=Data.Owner as TDataForm;             // note: change form reference!
  end else
  begin
    EndLine:=Form.Worksheet.Selection.Bottom;        // initialize cells range
    if EndLine=Form.Worksheet.RowCount-1
    then EndLine:=Form.Worksheet.RowCount-2;
    EndLine:=EndLine-Form.Worksheet.FixedRows;
    BegLine:=Form.Worksheet.Selection.Top-Form.Worksheet.FixedRows;
    XCol:=Form.XColumn; YCol:=Form.YColumn; Data:=Form.Container;
  end;
  Form.SaveUndoInfo(Action);
  with MainForm do
  begin
    StatusImageList.Tag:=8; DisplayHint(strWorking);             // show hints
    ProgressBar.Visible:=true; UpdatePicture(3); J:=-1;
  end;
  try
    try
      for I:=BegLine to EndLine do
      begin
        if EndLine>BegLine then MainForm.ProgressBar.Position:=   // check /0!
        Round((I-BegLine)/(EndLine-BegLine)*100); J:=I;       // show progress
        D:=Form.Container.Items[I]; ProcessItem(I,D);          // process item
      end;
    finally
      MainForm.StatusImageList.Tag:=0; MainForm.DisplayHint(''); // hide hints
      if Modified then Form.Container.Modified:=true;      // update container
    end;
  except
    if (J>=0) and (J<Data.Items.Count) then with Form do
    begin                                   // scroll worksheet to error point
      Show; if WindowState=wsMinimized then WindowState:=wsNormal; Scroll(J);
    end;
    MessageDlg(Format(strProcessError,[J]),mtError,[mbCancel],0); raise;
  end;
end;

{ TSwapColumns }

procedure TSwapColumns.ProcessItem(I: integer; D: TRealData);
var R: TReal;
begin
  R:=D.RData[XCol]; D.RData[XCol]:=D.RData[YCol]; D.RData[YCol]:=R;
end;

{ TParse }

procedure TParse.ProcessItem(I: integer; D: TRealData);
var K,N: integer; A: TReal;
begin
  N:=D.GetRData(CalculatorForm.Parameters);
  for K:=N+1 to MaxCols do CalculatorForm.Parameters[K]:=1.0;     // all valid
  with CalculatorForm do begin X:=Parameters[XCol]; Y:=Parameters[YCol]; end;
  A:=CalculatorForm.Parser.Parse(Macro);
  if InsCol then D.InsItem(A) else D.SetItem(YCol,A);
end;

{ TIntegration }

procedure TIntegration.ProcessItem(I: integer; D: TRealData);
var X1,X2,Y1,Y2: TReal; D1: TRealData;
const R: TReal=0;                                       {integral accumulator}
begin
  if I=BegLine then begin R:=0; D.InsItem(R); Exit; end;      {first value=0!}
  X2:=D.RData[XCol]; Y2:=D.RData[YCol];
  D1:=Data.Items[I-1]; X1:=D1.RData[XCol]; Y1:=D1.RData[YCol]; {get prev. pnt}
  R:=R+(Y1+Y2)/2*(X2-X1);             {accumulate integral (trapezoid method)}
  D.InsItem(R);
end;

{ TInterpolation }

procedure TInterpolation.ProcessItem(I: integer; D: TRealData);
var X1, X2, Y1, Y2: TReal; D1: TRealData;
const XI: TReal=0;
begin
  if I=BegLine then     {copy first element. Insert to END of current window!}
  begin
    D1:=TRealData(Data.InitItem); XI:=D.GetItem(XCol);        {remember 1st X}
    D1.InsItem(XI); D1.InsItem(D.GetItem(YCol)); Data.Items.Add(D1); Exit;{!!}
  end;
  D1:=Data.Items[I-1]; X1:=D1.RData[XCol]; X2:=D.RData[XCol];
  if X1=X2 then Exit;         {current interval is empty (duplicate X points)}
  Y1:=D1.RData[YCol]; Y2:=D.RData[YCol];
  while XI+Step<=X2 do
  begin
    XI:=XI+Step; D1:=TRealData(Data.InitItem); {make new element with next XI}
    D1.InsItem(XI); D1.InsItem(LineInterpolate(X1,X2,Y1,Y2,XI));
    Data.Items.Add(D1);
  end;
end;

{ TDifferentiation }

procedure TDifferentiation.ProcessItem(I: integer; D: TRealData);
var D1,D3: TRealData; RA,RB,X1,X2,X3,Y1,Y2,Y3: TReal;
begin                                 {if can't derive, value simply not set!}
  if (I=BegLine) or (I=EndLine) then Exit; {1st & last points processed later}
  D1:=Data.Items[I-1]; D3:=Data.Items[I+1];
  X1:=D1.RData[XCol]; Y1:=D1.RData[YCol];
  X2:=D.RData[XCol]; Y2:=D.RData[YCol];
  X3:=D3.RData[XCol]; Y3:=D3.RData[YCol];
  if (X1<>X2) and (X2<>X3) and (X1<>X3) then           {prevent division by 0}
  begin
    RA:=((Y1-Y2)/(X1-X2)-(Y2-Y3)/(X2-X3))/(X1-X3);  {calculate parab. coef-s:}
    RB:=((Y1-Y2)/(X1-X2)+(Y2-Y3)/(X2-X3)-RA*(X1+2*X2+X3))/2;
    D.InsItem(2*RA*X2+RB);
    if I=BegLine+1 then D1.InsItem(2*RA*X1+RB);  {values in 1st & last points}
    if I=EndLine-1 then D3.InsItem(2*RA*X3+RB);   {calculated by coefficients}
  end;                                               {in 2nd & last-1 points!}
end;

{ TSmooth }

procedure TSmooth.ProcessItem(I: integer; D: TRealData);
var bfXY,bfY,bfX,bfX2,Aa,Bb,X,Y: TReal; Count,N1,N2: integer; Dd: TRealData;
begin
  if I<BegLine+HW then N1:=BegLine else N1:=I-HW;           {calc BF vicinity}
  if I>EndLine-HW then N2:=EndLine else N2:=I+HW;
  bfX:=0; bfX2:=0; bfY:=0; bfXY:=0;
  for Count:=N1 to N2 do                             {calculate best fit sums}
  begin
    Dd:=Data.Items[Count]; X:=Dd.RData[XCol]; Y:=Dd.RData[YCol];
    bfX:=bfX+X; bfY:=bfY+Y; bfXY:=bfXY+X*Y; bfX2:=bfX2+X*X;
  end;
  Aa:=(N2-N1+1)*bfX2-bfX*bfX; if Aa<>0 then Aa:=((N2-N1+1)*bfXY-bfX*bfY)/Aa;
  Bb:=(bfY-Aa*bfX)/(N2-N1+1);                             {and best fit coefs}
  X:=D.RData[XCol]; if Derive then D.InsItem(Aa) else D.InsItem(Aa*X+Bb);
end;

end.
